// Datei GleitRech09.java
/* ------------------------------------------------------------------------
Werte des Typs float mit den Gleichheitsoperatoren == oder != zu
vergleichen ist haeufig nicht sinnvoll:

Da die Ergebnisse von Berechnungen mit float-Werten in aller Regel
mit einem Fehler behaftet sind, ist es haeufig sinnvoll, zwei solche
Ergebnisse (d.h. float-Werte) auch dann als gleich zu betrachten,
wenn sie sich geringfuegig unterscheiden (je nach Anwendung z.B. um
1 Prozent oder um 0.03 Prozent etc,). Die Funktion sindUngefaehrGleich
vergleicht zwei float-Werte auf diese Weise (der Prozentsatz erlaubter
Unterschiede muss als dritter Parameter angegeben werden).

Zum Typ float gehoeren insgesamt 2 hoch 32 (d.h. etwa 4.3 Milliarden)
Werte. 2 hoch 23 dieser float-Werte (d.h. etwa 8 Millionen oder etwa
0.5 Promille aller float-Werte) sind sogenannte NaN-Werte (Not a Number
values, oder "Unzahlen"). Alle anderen float-Werte repraesentieren Zahlen.

Vergleicht man einen NaN-Wert mit dem Gleichheitsoperator == mit sich
selbst (oder mit irgendeinem anderen float-Wert), so ist das Ergebnis
immer gleich false. Die Vergleichsfunktion sindExaktGleich (siehe unten)
liefert in solchen Faellen das Ergebnis true.

Die Vergleichsfunktion sindZiemlichGleich liefert auch dann true, wenn
ihre beiden Parameter verschiedene NaN-Werte sind.

Die Funktion Float.floatToRawIntBits hat einen float-Parameter g.
Sie liefert die Bitkombination von g als int-Wert.

Die Funktion Float.floatToIntBits leistet ganz Aehnliches, bildet aber
alle NaN-Werte auf dieselbe Bitkombination ab.
------------------------------------------------------------------------ */
class GleitRech09 {
   // ---------------------------------------------------------------------
   static public boolean sindExaktGleich(float g1, float g2) {
      // Liefert true, wenn g1 und g2 bitweise gleich sind, und sonst false.
      // Liefert (anders als der Gleichheitsoperator ==) insbesondere auch
      // dann true, wenn g1 und g2 bitweise gleiche NaN-Werte enthalten!

      int n1 = Float.floatToRawIntBits(g1);
      int n2 = Float.floatToRawIntBits(g2);
      return n1 == n2;
   } // sindExaktGleich
   // ---------------------------------------------------------------------
   static public boolean sindZiemlichGleich(float g1, float g2) {
      // Liefert true, wenn g1 und g2 (gleiche oder verschiedene) NaN-Werte
      // enthalten oder gleiche float-Zahlen, und liefert sonst false.

      int n1 = Float.floatToIntBits(g1);
      int n2 = Float.floatToIntBits(g2);
      return n1 == n2;
   } // sindZiemlichGleich
   // ---------------------------------------------------------------------
   static public boolean sindUngefaehrGleich(float g1, float g2, float p) {
      // Liefert true, wenn sich g1 und g2 um hoechstens p Prozent vonein-
      // ander unterscheiden, und liefert sonst false.
      // Genauer: Liefert true, wenn die Differenz g1-g2 hoechstens
      // p Prozent von g1 und hoechstens p Prozent von g2 ausmacht,
      // und liefert sonst false.

      float delta     = Math.abs(g1 - g2);
      float maxDelta1 = Math.abs(g1 * p / 100.0F); // p Prozent von g1
      float maxDelta2 = Math.abs(g2 * p / 100.0F); // p Prozent von g2

      return delta <= maxDelta1 && delta <= maxDelta2;
   } // sindUngefaehrGleich
   // ---------------------------------------------------------------------
   static public void main(String[] sonja) {
      // Die Methoden sindExaktGleich und sindUngefaehrGleich werden
      // ein bisschen getestet.

      // Die Variablen nan1 und nan2 werden mit zwei unterschiedlichen
      // NaN-Werten initialisiert:
      float g0   = 0.0F;
      float nan1 = g0 / g0;                     // Ein NaN-Wert
      int   bits = Float.floatToRawIntBits(nan1) + 5 ;
      float nan2 = Float.intBitsToFloat(bits);  // Ein anderer NaN-Wert

       // Die Variablen g1 und g2 werden mit fast-gleichen Werten
       // initialisiert:
      float g1   = 200.0F;  // g1 und g2 unterscheiden sich nur um
      float g2   = 200.3F;  // etwa 0.15 Prozent (1.5 Promille)

      pln("GleitRech09: Jetzt geht es los!");
      pln("-----------------------------------------------------");
      pln("A                                   nan1 : " + nan1);
      pln("B                                   nan2 : " + nan2);
      pln("C           Float.floatToRawIntBits(nan1): " +
                       Float.floatToRawIntBits(nan1));
      pln("D           Float.floatToRawIntBits(nan2): " +
                       Float.floatToRawIntBits(nan2));
      pln("E           Float.floatToIntBits   (nan1): " +
                       Float.floatToIntBits   (nan1));
      pln("F           Float.floatToIntBits   (nan2): " +
                       Float.floatToIntBits   (nan2));
      pln("G                                     g1 : " + g1);
      pln("H                                     g2 : " + g2);
      pln("-----------------------------------------------------");
      pln("I                            (nan1==nan1): " +
                                        (nan1==nan1));
      pln("J                            (nan1==nan2): " +
                                        (nan1==nan2));
      pln("-----------------------------------------------------");
      pln("K             sindExaktGleich(nan1, nan1): " +
                         sindExaktGleich(nan1, nan1));
      pln("L             sindExaktGleich(nan1, nan2): " +
                         sindExaktGleich(nan1, nan2));
      pln("-----------------------------------------------------");
      pln("M          sindZiemlichGleich(nan1, nan1): " +
                      sindZiemlichGleich(nan1, nan1));
      pln("N          sindZiemlichGleich(nan1, nan2): " +
                      sindZiemlichGleich(nan1, nan2));
      pln("-----------------------------------------------------");
      pln("O    sindUngefaehrGleich(g1, g2,    0.1F): " +
                sindUngefaehrGleich(g1, g2,    0.1F));
      pln("P    sindUngefaehrGleich(g1, g2,    0.2F): " +
                sindUngefaehrGleich(g1, g2,    0.2F));
      pln("-----------------------------------------------------");
      pln("Q sindUngefaehrGleich(nan1, nan1,  10.0F): " +
             sindUngefaehrGleich(nan1, nan1,  10.0F));
      pln("R sindUngefaehrGleich(nan1, nan2, 100.0F): " +
             sindUngefaehrGleich(nan1, nan2, 100.0F));
      pln("-----------------------------------------------------");
      pln("GleitRech09: Das war's erstmal!");
   } // main
   // ---------------------------------------------------------------------
   // Eine Methode mit einem kurzen Namen:
   static void pln(Object ob) {System.out.println(ob);}
   // ---------------------------------------------------------------------
} // class GleitRech09
/* ------------------------------------------------------------------------
Ausgabe des Programms GleitRech09:

GleitRech09: Jetzt geht es los!
-----------------------------------------------------
A                                   nan1 : NaN
B                                   nan2 : NaN
C           Float.floatToRawIntBits(nan1): -4194304
D           Float.floatToRawIntBits(nan2): -4194299
E           Float.floatToIntBits   (nan1): 2143289344
F           Float.floatToIntBits   (nan2): 2143289344
G                                     g1 : 200.0
H                                     g2 : 200.3
-----------------------------------------------------
I                            (nan1==nan1): false
J                            (nan1==nan2): false
-----------------------------------------------------
K             sindExaktGleich(nan1, nan1): true
L             sindExaktGleich(nan1, nan2): false
-----------------------------------------------------
M          sindZiemlichGleich(nan1, nan1): true
N          sindZiemlichGleich(nan1, nan2): true
-----------------------------------------------------
O    sindUngefaehrGleich(g1, g2,    0.1F): false
P    sindUngefaehrGleich(g1, g2,    0.2F): true
-----------------------------------------------------
Q sindUngefaehrGleich(nan1, nan1,  10.0F): false
R sindUngefaehrGleich(nan1, nan2, 100.0F): false
-----------------------------------------------------
GleitRech09: Das war's erstmal!
------------------------------------------------------------------------ */