// Datei GleitRech10.java
/* ------------------------------------------------------------------------
Werte des Typs double mit den Gleichheitsoperatoren == oder != zu
vergleichen ist haeufig nicht sinnvoll:

Da die Ergebnisse von Berechnungen mit double-Werten in aller Regel
mit einem Fehler behaftet sind, ist es haeufig sinnvoll, zwei solche
Ergebnisse (d.h. double-Werte) auch dann als gleich zu betrachten,
wenn sie sich geringfuegig unterscheiden (je nach Anwendung z.B. um
1 Prozent oder um 0.5 Prozent etc,). Die Funktion sindUngefaehrGleich
vergleicht zwei double-Werte auf diese Weise (der Prozentsatz erlaubter
Unterschiede muss als dritter Parameter angegeben werden).

Zum Typ double gehoeren insgesamt 2 hoch 32 (d.h. etwa 4.3 Milliarden)
Werte. 2 hoch 23 dieser double-Werte (d.h. etwa 8 Millionen oder etwa
0.5 Promille aller double-Werte) sind sogenannte NaN-Werte (Not a Number
values, oder "Unzahlen"). Alle anderen double-Werte repraesentieren Zahlen.

Vergleicht man einen NaN-Wert mit dem Gleichheitsoperator == mit sich
selbst (oder mit irgendeinem anderen double-Wert), so ist das Ergebnis
immer gleich false. Die Vergleichsfunktion sindExaktGleich (siehe unten)
liefert in solchen Faellen das Ergebnis true.

Die Vergleichsfunktion sindZiemlichGleich liefert auch dann true, wenn
ihre beiden Parameter verschiedene NaN-Werte sind.

Die Funktion Double.doubleToRawLongBits hat einen double-Parameter g.
Sie liefert die Bitkombination von g als Long-Wert.

Die Funktion Double.doubleToLongBits leistet ganz Aehnliches, bildet aber
alle NaN-Werte auf dieselbe Bitkombination ab.
------------------------------------------------------------------------ */
class GleitRech10 {
   // ---------------------------------------------------------------------
   static public boolean sindExaktGleich(double g1, double g2) {
      // Liefert true, wenn g1 und g2 bitweise gleich sind, und sonst false.
      // Liefert (anders als der Gleichheitsoperator ==) insbesondere auch
      // dann true, wenn g1 und g2 bitweise gleiche NaN-Werte enthalten!

      long n1 = Double.doubleToRawLongBits(g1);
      long n2 = Double.doubleToRawLongBits(g2);
      return n1 == n2;
   } // sindExaktGleich
   // ---------------------------------------------------------------------
   static public boolean sindZiemlichGleich(double g1, double g2) {
      // Liefert true, wenn g1 und g2 bitweise gleich sind, und sonst false.
      // Liefert (anders als der Gleichheitsoperator ==) insbesondere auch
      // dann true, wenn g1 und g2 bitweise gleiche NaN-Werte enthalten!

      long n1 = Double.doubleToLongBits(g1);
      long n2 = Double.doubleToLongBits(g2);
      return n1 == n2;
   } // sindZiemlichGleich
   // ---------------------------------------------------------------------
   static public boolean sindUngefaehrGleich(double g1, double g2, double p) {
      // Liefert true, wenn sich g1 und g2 um hoechstens p Prozent vonein-
      // ander unterscheiden, und liefert sonst false.
      // Genauer: Liefert true, wenn die Differenz g1-g2 hoechstens
      // p Prozent von g1 und hoechstens p Prozent von g2 ausmacht,
      // und liefert sonst false.

      double delta     = Math.abs(g1 - g2);
      double maxDelta1 = Math.abs(g1 * p / 100.0); // p Prozent von g1
      double maxDelta2 = Math.abs(g2 * p / 100.0); // p Prozent von g2

      return delta <= maxDelta1 && delta <= maxDelta2;
   } // sindUngefaehrGleich
   // ---------------------------------------------------------------------
   static public void main(String[] sonja) {
      // Die Methoden sindExaktGleich und sindUngefaehrGleich werden
      // ein bisschen getestet.

      // Die Variablen nan1 und nan2 werden mit zwei unterschiedlichen
      // NaN-Werten initialisiert:
      double g0   = 0.0F;
      double nan1 = g0 / g0;                     // Ein NaN-Wert
      long   bits = Double.doubleToRawLongBits(nan1) + 9 ;
      double nan2 = Double.longBitsToDouble(bits);  // Ein anderer NaN-Wert

       // Die Variablen g1 und g2 werden mit fast-gleichen Werten
       // initialisiert:
      double g1   = 200.0F;  // g1 und g2 unterscheiden sich nur um
      double g2   = 200.3F;  // etwa 0.15 Prozent (1.5 Promille)

      pln("GleitRech10: Jetzt geht es los!");
      pln("-------------------------------------------------------------");
      pln("A                                  nan1 : " + nan1);
      pln("B                                  nan2 : " + nan2);
      pln("C       Double.doubleToRawLongBits(nan1): " +
                   Double.doubleToRawLongBits(nan1));
      pln("D       Double.doubleToRawLongBits(nan2): " +
                   Double.doubleToRawLongBits(nan2));
      pln("E       Double.doubleToLongBits   (nan1): " +
                   Double.doubleToLongBits   (nan1));
      pln("F       Double.doubleToLongBits   (nan2): " +
                   Double.doubleToLongBits   (nan2));
      pln("G                                    g1 : " + g1);
      pln("H                                    g2 : " + g2);
      pln("-------------------------------------------------------------");
      pln("I                           (nan1==nan1): " +
                                       (nan1==nan1));
      pln("J                           (nan1==nan2): " +
                                       (nan1==nan2));
      pln("-------------------------------------------------------------");
      pln("K            sindExaktGleich(nan1, nan1): " +
                        sindExaktGleich(nan1, nan1));
      pln("L            sindExaktGleich(nan1, nan2): " +
                        sindExaktGleich(nan1, nan2));
      pln("-------------------------------------------------------------");
      pln("M         sindZiemlichGleich(nan1, nan1): " +
                     sindZiemlichGleich(nan1, nan1));
      pln("N         sindZiemlichGleich(nan1, nan2): " +
                     sindZiemlichGleich(nan1, nan2));
      pln("-------------------------------------------------------------");
      pln("O    sindUngefaehrGleich(g1, g2,    0.1): " +
                sindUngefaehrGleich(g1, g2,    0.1));
      pln("P    sindUngefaehrGleich(g1, g2,    0.2): " +
                sindUngefaehrGleich(g1, g2,    0.2));
      pln("-------------------------------------------------------------");
      pln("Q sindUngefaehrGleich(nan1, nan1,  10.0): " +
             sindUngefaehrGleich(nan1, nan1,  10.0));
      pln("R sindUngefaehrGleich(nan1, nan2, 100.0): " +
             sindUngefaehrGleich(nan1, nan2, 100.0));
      pln("-------------------------------------------------------------");
      pln("GleitRech10: Das war's erstmal!");
   } // main
   // ---------------------------------------------------------------------
   // Eine Methode mit einem kurzen Namen:
   static void pln(Object ob) {System.out.println(ob);}
   // ---------------------------------------------------------------------
} // class GleitRech10
/* ------------------------------------------------------------------------
Ausgabe des Programms GleitRech10:

GleitRech10: Jetzt geht es los!
-------------------------------------------------------------
A                                  nan1 : NaN
B                                  nan2 : NaN
C       Double.doubleToRawLongBits(nan1): -2251799813685248
D       Double.doubleToRawLongBits(nan2): -2251799813685239
E       Double.doubleToLongBits   (nan1): 9221120237041090560
F       Double.doubleToLongBits   (nan2): 9221120237041090560
G                                    g1 : 200.0
H                                    g2 : 200.3000030517578
-------------------------------------------------------------
I                           (nan1==nan1): false
J                           (nan1==nan2): false
-------------------------------------------------------------
K            sindExaktGleich(nan1, nan1): true
L            sindExaktGleich(nan1, nan2): false
-------------------------------------------------------------
M         sindZiemlichGleich(nan1, nan1): true
N         sindZiemlichGleich(nan1, nan2): true
-------------------------------------------------------------
O    sindUngefaehrGleich(g1, g2,    0.1): false
P    sindUngefaehrGleich(g1, g2,    0.2): true
-------------------------------------------------------------
Q sindUngefaehrGleich(nan1, nan1,  10.0): false
R sindUngefaehrGleich(nan1, nan2, 100.0): false
-------------------------------------------------------------
GleitRech10: Das war's erstmal!
------------------------------------------------------------------------ */