// Datei Objects.java
/* ------------------------------------------------------------------------
Mit der Methode Arrays.toString kann man eine einstufige Reihungen von
Objekten in einen String umwandeln. Eine Anwendung auf Reihungen mit
primitiven Komponenten ist verboten und eine Anwendung auf mehrstufige
Reihungen erzeugt meist unbefriedigende Ergebnisse.
Die Methode Arrays.deepToString darf man nur auf (ein- oder mehrstufige)
Reihungen von Objekten anwenden, nicht auf Reihungen mit primitiven
Komponenten. Der erzeugte String kann sehr lang sein und es ist dann
nicht leicht, ihn "uebersichtlich und lesbar" auszugeben.

Mit der hier vereinbarten Methode Objects.toString kann man ein beliebiges
Objekt in einen "einigermassen lesbaren" String umwandeln, der auch
Informationen ueber den Typ des Objekts enthaelt. Falls der String nicht
auf eine Zeile passt wird versucht, ihn durch Einfuegen von Zeilenwechseln
"lesbar" umzubrechen (das gelingt manchmal, aber nicht immer).
Das umzuwandelnde Objekt darf insbesondere auch eine ein- oder mehrstufige
Reihung (engl. an array or nested array) sein, deren Komponenten beliebige
Objekte oder primitive Werte sind.
---------------------------------------------------------------------------
Die Werte der folgenden Klassenattribute beeinflussen den Zeilenumbruch.
Man kann sie mit entsprechenden set-Methoden veraendern.

zeilenLaenge    Die Laenge einer Ausgabezeile (z.B. 60 oder 80)
zeichenProERS   Zeichen pro Einrueckstufe     (z.B.  4 oder  3)

Einrueckstufen sind vor allem bei der Ausgabe von mehrdimensionalen
Reihungen wichtig. Ausserdem werden "zu lange Zeilen" automatisch
umgebrochen und die Folgezeilen werden eine Stufe mehr eingerueckt.
------------------------------------------------------------------------ */
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.Array;

class Objects {
   // ---------------------------------------------------------------------
   // Die oeffentliche Schnittstelle der Klasse (3 Methoden):
   static public String toString(Object ob) {
      // Wandelt ob in einen "einigermassen lesbaren" String um, der
      // moeglicherweise Zeilenwechselzeichen enthaelt. Als Param. ob
      // ist jedes Objekt erlaubt, auch eine beliebige Reihung (array).

      // Umgewandelt wird mit den Methoden print und println eines
      // PrintWriter-Objekts:
      StringWriter sw = new StringWriter();
      PrintWriter  pw = new PrintWriter(sw);

      // Jetzt wird ob in einen String umgewandelt:
      printObInit(ob, 0, pw);
      return sw.toString();
   } // toString

   static public void setZeilenLaenge(int zL) {
      // Setzt die zeilenLaenge auf zL:
      zeilenLaenge = zL;
   } // setZeilenLaenge

   static public void setZeichenProERS(int zPE) {
      // Setzt die zeichenProERS (Zeichen pro Einrueckstufe) auf zPE:
      zeichenProERS = zPE;
   } // setZeichenProERS

   // Ende der oeffentlichen Schnittstelle der Klasse
   // ---------------------------------------------------------------------
   // Der "globale Zustand" dieser Klasse:
   static private int zeilenLaenge     = 60;
   static private int zeichenProERS    = 4;  // Pro EinRueckStufe
   static private int ersteFreieSpalte;      // Init. siehe printObInit
   // ---------------------------------------------------------------------
   static private void printObInit(Object ob, int erStufe, PrintWriter pw) {
      // Wird *nicht* rekursiv aufgerufen. Initialisiert ersteFreieSpalte:
      ersteFreieSpalte = 0;
      printOb(ob, erStufe, pw);
   } // printObInit
   // ---------------------------------------------------------------------
   static private void printOb(Object ob, int erStufe, PrintWriter pw) {
      // Wird evtl. rekursiv aufgerufen.
      // Gibt das Objekt ob mit Einrueckstufe erStufe zum Strom pw
      // aus. Die Einrueckstufe ist wichtig, falls ob eine Komponente
      // einer (mehrstufigen) Reihung ist.

      // Falls ob auf kein Objekt zeigt:
      if (ob == null) {
         gibAus("null-Referenz", erStufe, pw);
         return;
      }

      // Von welchem Typ ist das auszugebende Objekt ob?
      Class  obKlasse      = ob.getClass();
      String obKlassenName = obKlasse.getName();

      // Falls der Klassen-Name mit "java.lang." beginnt, wird dieser
      // Anfang abgeschnitten ("stoerendes Rauschen entfernen"):
      if (obKlassenName.startsWith("java.lang.")) {
         obKlassenName = obKlassenName.substring(10);
      }

      // Falls das Objekt ob eine Reihung ist, wird es mit der
      // Methode printR ausgegeben:
      if (obKlasse.isArray()) {
         printR(ob, obKlasse, obKlassenName, erStufe, pw);
         return;
      }

      // Falls ob keine Reihung (sondern "ein normales Objekt") ist:
      gibAus("Typ: " + obKlassenName + ", Wert: " + ob, erStufe, pw);
   } // printOb
   // ---------------------------------------------------------------------
   static private void printR
      (Object      r,            // Die auszugebende Reihung
       Class       rKlasse,      // Das Klassen-Objekt der Klasse von r
       String      rKlassenName, // Der Name dieser Klasse
       int         erStufe,      // Die aktuelle Einrueckstufe
       PrintWriter pw)           // Dahin wird ausgegeben
   {
      // Verlaesst sich darauf, dass r eine Reihung ist. Gibt r zum
      // Strom pw aus.
      // ------------------------------------------------------------------
      // Die von rKlasse.getName() gelieferten Klassennamen sind nicht
      // besonders lesbar. Deshalb werden sie hier "uebersetzt":
      rKlassenName = machLesbar(rKlassenName, pw);
      int laenge = Array.getLength(r);
      // Allgemeine Informationen ueber die Reihung r ausgeben:
      gibAus("Typ: " + rKlassenName +
         ", Laenge: " + laenge, erStufe, pw);
      if (laenge == 0) return;

      // Falls die Reihung r Komponenten eines primitiven Typs besitzt,
      // wird sie mit der Methode printRP ausgegeben:
      Class kKlasse = rKlasse.getComponentType();
      if (kKlasse.isPrimitive()) {
         printRP(r, laenge, rKlassenName, erStufe, pw);
         return;
      }

      // r ist also eine Reihung mit nicht-primitiven Komponenten. Diese
      // Komponenten werden jetzt ausgegeben. Fuer jede Komponente wird
      // eine neue Zeile begonnen und die Einrueckstufe wird um 1 erhoeht.
      for (int i=0; i<laenge; i++) {
         int neueEinrueckung = erStufe + 1;
         neueZeile(neueEinrueckung, pw);
         gibAus("[" + i + "] ", neueEinrueckung, pw);
         Object komp = Array.get(r, i);
         printOb(komp, neueEinrueckung, pw); // Ein indirekt rekursiver Aufruf!
      }
   } // printR
   // ---------------------------------------------------------------------
   static private void printRP
      (Object r,            // Die auszugebende Reihung
       int laenge,          // Ihre Laenge (Anzahl Komponenten)
       String rKlassenName, // Der Name ihres Typs
       int erStufe,         // Die aktuelle Einrueckstufe
       PrintWriter pw)      // Dahin wird ausgegeben
   {
      // Verlaesst sich darauf, dass r eine nicht-leere Reihung mit
      // Komponenten eines primitiven Typs (byte, char, short, ... boolean)
      // ist, Gibt r zum Strom pw aus (falls moeglich alle Komponenten
      // auf einer Zeile):

      neueZeile(erStufe+1, pw);

      // Die Komponente mit Index 0 ist ein Sonderfall:
      StringBuffer sb = new StringBuffer("{" + Array.get(r, 0));

      // Alle anderen Komponenten werden gleich behandelt:
      for (int i=1; i< laenge; i++) {
         sb.append(", " + Array.get(r, i));
      }
      sb.append("}");
      gibAus(sb.toString(), erStufe, pw);
   } // printRP
   // ---------------------------------------------------------------------
   static private void neueZeile(int erStufe, PrintWriter pw) {
      // Beginnt eine neue Zeile und gibt so viele Blanks aus, wie der
      // Einrueckstufe erStufe entsprechen:
      pw.println();
      final int einRueckung = zeichenProERS * erStufe;
      for (int i=0; i< einRueckung; i++) pw.print(" ");
      ersteFreieSpalte = einRueckung;
   } // neueZeile
   // ---------------------------------------------------------------------
   static private void gibAus(String s, int erStufe, PrintWriter pw) {
      // Gibt s zum Strom pw aus. Falls s nicht ganz auf die
      // aktuelle Zeile passt, werden Folgezeilen ausgegeben. Dabei
      // wird die Einrueckstufe erStufe beachtet.

      // Wieviele Spalten der aktuellen Zeile sind noch frei?
      int nochFrei = zeilenLaenge - ersteFreieSpalte;
      // Wie lang kann eine Folgezeile sein (falls eine noetig wird):
      int folgeZL = zeilenLaenge - zeichenProERS * (erStufe+1);

      // Falls die Einrueckung so tief ist, dass Folgezeilen zu kurz
      // werden wuerden:
      if (folgeZL < 10) folgeZL += zeichenProERS;

      // Falls s ganz auf die aktuelle Zeile passt (ohne Folgezeilen):
      if (s.length() <= nochFrei) {
         pw.print(s);
         ersteFreieSpalte += s.length();
         return;
      }

      // Die erste von mehreren Zeilen ausgeben:
      pw.print(s.substring(0, nochFrei));
      int fertig = nochFrei;

      // Die uebrigen Zeilen ausgeben:
      while (fertig <s.length()) {
         neueZeile(erStufe+1, pw);
         int bis = Math.min(fertig+folgeZL, s.length());
         pw.print(s.substring(fertig, bis));
         fertig = bis;
      }
      ersteFreieSpalte = fertig + zeichenProERS * erStufe;
   } // gibAus
   // ---------------------------------------------------------------------
   static public String machLesbar(String alterTypName, PrintWriter pw) {
      // Wandelt einen von getName() gelieferten Klassen-Namen in eine
      // lesbarere Form um (z.B. wird "[I" in "int[]" umgewandelt und
      // [[[Ljava.lang.String;" in "String[][][]").

      // Die Anzahl der '['-Zeichen am Anfang von alterTypName zaehlen.
      // Diese Anzahl gibt die Dimensionen eines Reihungstyps an.
      int anZahl = 0;
      for (int i=0; i<alterTypName.length(); i++) {
         if (alterTypName.charAt(i) == '[') {
            anZahl++;
         } else {
            break; // Beim ersten Zeichen ungleich '[' aufhoeren
         }
      } // for

      // Falls der alte Typname nicht der Name eines Reihungstyps ist,
      // wird er unveraendert zurueckgeliefert:
      if (anZahl == 0) return alterTypName;

      // Jetzt ist der alte Typname mit Sicherhiet der Name eines Reihungs-
      // typs und wird in einen neuen ("besseren") Namen umgewandelt:
      StringBuffer neuerName = new StringBuffer();
      switch (alterTypName.charAt(anZahl)) {
         case 'B': neuerName.append("byte");    break;
         case 'C': neuerName.append("char");    break;
         case 'S': neuerName.append("short");   break;
         case 'I': neuerName.append("int");     break;
         case 'J': neuerName.append("long");    break;
         case 'F': neuerName.append("float");   break;
         case 'D': neuerName.append("double");  break;
         case 'Z': neuerName.append("boolean"); break;
         case 'L': { // Ein Reihungstyp mit nicht-primitivem Komp.Typ.
            int von = anZahl+1;                // die '['-Zeichen weg
            int bis = alterTypName.length()-1; // abschliessendes ';' weg
            neuerName.append(alterTypName.substring(von, bis));

            // Falls der Klassen-Name mit "java.lang." beginnt, wird dieser
            // Anfang abgeschnitten ("stoerendes Rauschen entfernen"):
            String s = neuerName.toString();
            if (s.startsWith("java.lang.")) {
               s = s.substring(10);
               neuerName = new StringBuffer(s);
            }
            break;
         }
         default: {
            // Dieser Fall darf eigentlich nicht auftreten:
            gibAus("machLesbar: Falscher Klassenname " + alterTypName, 0, pw);
         }
      } // switch

      // An den neuen Typnamen die richtige Anzahl von "[]" anhaengen:
      for (int i=1; i<=anZahl; i++) {
         neuerName.append("[]");
      }
      return neuerName.toString();
   } // machLesbar
   // ---------------------------------------------------------------------
   // Mehrere Methoden mit kurzen Namen:
   static void pln(Object ob) {System.out.println(ob);}
   static void p  (Object ob) {System.out.print  (ob);}
   static void pln()          {System.out.println();  }
   // ---------------------------------------------------------------------
   // Eine Objektmethode (ist hier nicht wichtig, kann auch entfernt werden)
   public String toString() {
      return "Dieses Objects-Objekt sollte nicht mit " +
             "einem Object-Objekt verwechselt werden!!!";
   } // toString
   // =====================================================================
   static public void main(String [] sonja) {
      // Gibt verschiedene Objekte (insbesonder ein- und mehrstufige
      // Reihung aus (zum Testen der Methode toString):

      pln("Objects: Jetzt geht es los!");

      // "Normale Objekte" (die keine Reihungen sind):
      String       s0     = null;
      String       s1     = "Hallo Susi! Dies ist ein ziemlich langer "   +
                            "String, der beim Ausgeben wahrscheinlich "   +
                            "mind. einen Zeilenwechsel erzwingt!";
      StringBuffer sb1    = new StringBuffer("Wie geht es? Ach danke, "   +
                            "mir geht es eigentlch recht gut, bis auf "   +
                            " die anstehenden  PR1-Klausuren, die "       +
                            "machen mir Sorgen!");

      // Reihungen mit Komponenten primitiver Typen. Einige Reihungen sind
      // so lang, dass sie kaum auf einer Zeile ausgegeben werden koennen.
      byte   [] byteR1    = {17, -32, 127, -128, 11, 22, 33, 44, 55,
                             66, 77, 88, 99, 11, 22, 33, 44, 55, 66,
                             77, 88, 99, 11, 22, 33, 44, 55, 66, 77};
      char   [] charR1    = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                             'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q',
                             'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z'};
      short  [] shortR1   = {+32767, -32768, 555};
      int    [] intR1     = {7, 5, 3, 1};
      long   [] longR1    = {77, 55, 33, 11};
      float  [] floatR1   = {2.2f, 2.4f, 2.6f, 3.7f};
      double [] doubleR1  = {8.1d, 9.3d, 5.7d, 3.3d};
      boolean[] booleanR1 = {false, false, true, false, true};

      // Eine eindimensionale Reihungen mit nicht-primitiven Komponenten:
      String     lang1     = "Hallo ... Hallo ... Hallo ... Hallo ... " +
                             "Hallo ... Hallo ... Hallo ... Hallo ... " +
                             "Hallo ... Hallo ... Hallo ... Hallo ... " +
                             "Hallo ... Hallo ... Hallo ... Hallo ... ";
      String[]   stringR1  = {lang1, "How are you?", "Fine, thanks!"};
      // Mehrdimensionale Reihungen mit primitiven und mit nicht-primitiven
      // Komponenten:
      int[][]    intRR1    = {{10, 11, 12}, {20, 21}, {30, 31, 32, 33}};
      String[][] stringRR1 = {{"A0", "A1", "A2"}, {"B0", "B1"},
                              {"C0", "C1", "C2", "C3"}};
      // Eine Reihung, die Reihungen, andere Objekte und null-Referenzen
      // enthaelt:
      Object[] obR1 = {new int[]{40, 41, 42}, null, null, new Objects(),
                       new String("Hello!"), stringRR1,  new Integer(17)};

      // Jetzt werden all diese Objekte mit der obigen Methode toString
      // in Strings umgewandelt und ausgegeben:
      pln("-------------------------------------------------------------");
      pln("A " + toString(s0       ) + "\n"); // Eine null-Referenz
      pln("B " + toString(s1       ) + "\n"); // Ein String
      pln("C " + toString(sb1      )       ); // Ein StringBuffer
      pln("-------------------------------------------------------------");
                                              // Eine Reihung von
      pln("D " + toString(byteR1   ) + "\n"); // byte
      pln("E " + toString(charR1   ) + "\n"); // char
      pln("F " + toString(shortR1  ) + "\n"); // short
      pln("G " + toString(intR1    ) + "\n"); // int
      pln("H " + toString(longR1   ) + "\n"); // long
      pln("I " + toString(floatR1  ) + "\n"); // float
      pln("J " + toString(doubleR1 ) + "\n"); // double
      pln("K " + toString(booleanR1)       ); // boolean
      pln("-------------------------------------------------------------");
      pln("L " + toString(stringR1 ));        // Strings
      pln("-------------------------------------------------------------");
      pln("M " + toString(intRR1   ));        // Reihungen von int
      pln("-------------------------------------------------------------");
      pln("N " + toString(stringRR1));        // Reihungen von String
      pln("-------------------------------------------------------------");
      pln("O " + toString(obR1     ));        // diversen Objekten
      pln("-------------------------------------------------------------");
      pln("Objects: Das war's erstmal!");
   } // main
   // ---------------------------------------------------------------------
} // class Objects
/* ------------------------------------------------------------------------
Ausgabe des Programms Objects:

Objects: Jetzt geht es los!
-------------------------------------------------------------
A null-Referenz

B Typ: String, Wert: Hallo Susi! Dies ist ein ziemlich langer
    String, der beim Ausgeben wahrscheinlich mind. einen Zei
    lenwechsel erzwingt!

C Typ: StringBuffer, Wert: Wie geht es? Ach danke, mir geht es
     eigentlch recht gut, bis auf  die anstehenden  PR1-Klau
    suren, die machen mir Sorgen!
-------------------------------------------------------------
D Typ: byte[], Laenge: 29
    {17, -32, 127, -128, 11, 22, 33, 44, 55, 66, 77, 88, 99,
     11, 22, 33, 44, 55, 66, 77, 88, 99, 11, 22, 33, 44, 55,
     66, 77}

E Typ: char[], Laenge: 26
    {A, B, C, D, E, F, G, H, I, J, K, L, M, N, O, P, Q, R, S
    , T, U, V, W, X, Y, Z}

F Typ: short[], Laenge: 3
    {32767, -32768, 555}

G Typ: int[], Laenge: 4
    {7, 5, 3, 1}

H Typ: long[], Laenge: 4
    {77, 55, 33, 11}

I Typ: float[], Laenge: 4
    {2.2, 2.4, 2.6, 3.7}

J Typ: double[], Laenge: 4
    {8.1, 9.3, 5.7, 3.3}

K Typ: boolean[], Laenge: 5
    {false, false, true, false, true}
-------------------------------------------------------------
L Typ: String[], Laenge: 3
    [0] Typ: String, Wert: Hallo ... Hallo ... Hallo ... Hal
        lo ... Hallo ... Hallo ... Hallo ... Hallo ... Hallo
         ... Hallo ... Hallo ... Hallo ... Hallo ... Hallo .
        .. Hallo ... Hallo ...
    [1] Typ: String, Wert: How are you?
    [2] Typ: String, Wert: Fine, thanks!
-------------------------------------------------------------
M Typ: int[][], Laenge: 3
    [0] Typ: int[], Laenge: 3
        {10, 11, 12}
    [1] Typ: int[], Laenge: 2
        {20, 21}
    [2] Typ: int[], Laenge: 4
        {30, 31, 32, 33}
-------------------------------------------------------------
N Typ: String[][], Laenge: 3
    [0] Typ: String[], Laenge: 3
        [0] Typ: String, Wert: A0
        [1] Typ: String, Wert: A1
        [2] Typ: String, Wert: A2
    [1] Typ: String[], Laenge: 2
        [0] Typ: String, Wert: B0
        [1] Typ: String, Wert: B1
    [2] Typ: String[], Laenge: 4
        [0] Typ: String, Wert: C0
        [1] Typ: String, Wert: C1
        [2] Typ: String, Wert: C2
        [3] Typ: String, Wert: C3
-------------------------------------------------------------
O Typ: Object[], Laenge: 7
    [0] Typ: int[], Laenge: 3
        {40, 41, 42}
    [1] null-Referenz
    [2] null-Referenz
    [3] Typ: Objects, Wert: Dieses Objects-Objekt sollte nic
        ht mit einem Object-Objekt verwechselt werden!!!
    [4] Typ: String, Wert: Hello!
    [5] Typ: String[][], Laenge: 3
        [0] Typ: String[], Laenge: 3
            [0] Typ: String, Wert: A0
            [1] Typ: String, Wert: A1
            [2] Typ: String, Wert: A2
        [1] Typ: String[], Laenge: 2
            [0] Typ: String, Wert: B0
            [1] Typ: String, Wert: B1
        [2] Typ: String[], Laenge: 4
            [0] Typ: String, Wert: C0
            [1] Typ: String, Wert: C1
            [2] Typ: String, Wert: C2
            [3] Typ: String, Wert: C3
    [6] Typ: Integer, Wert: 17
-------------------------------------------------------------
Objects: Das war's erstmal!
------------------------------------------------------------------------ */