// Datei Sammlungen07.java
/* ------------------------------------------------------------------------
Sammlungen eines parametrisierten Sammlungstyps (z. B. des Typs
Collection<Long> oder des Typs ArrayList<Long>) sind stark
typsicher, d. h. alle Typfehler werden schon bei der Uebergabe des
Programms ("zur Compilezeit") erkannt. Allerdings lehnt der Aus-
fuehrer nicht alle "typunsicheren" Programme ab: Vor der Verwendung
von rohen Typen warnt er nur.

Trotzdem koennen "boese Methoden in einer boesen Klasse" Objekte
falscher Typen in solche Sammlungen einfuegen, indem sie eine rohe
Variable auf die Sammlung zeigen lassen und die Warnungen des
Compilers vor rohen Typen einfach nicht beachten.

Die Methode Collections.checkedCollection liefert einem eine Sicht
(engl. a view) auf eine Sammlung, die gegen solche "boesen Methoden"
abgesichert ist: Ein Versuch, ein Objekt eines falschen Typs in
eine solche Sicht einer Sammlung einzufuegen loest eine Ausnahme
(des Typs ClassCastException) aus.

Diese Tatsachen werden durch dieses Programm illustriert.
------------------------------------------------------------------------ */
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections; // fuer checkedCollection
// ------------------------------------------------------------------------
class Boese {
   // Die Methode add fuegt ("boeserweise") in eine Sammlung des
   // Typs Collction<Integer> ein Double-Objekt ein.
   static Double dob = new Double(3.5);

   static public void add(Collection<Long> ci) {
      // Zugriff auf die Sammlung ci mit einer rohen Variablen cr:
      Collection cr = ci;
      cr.add(dob); // Vor diesem Befehl warnt der Compiler (aber
                   // die Mafia hoert ja nicht auf solche Warnungen :-)
   } // add
} // class Boese
// ------------------------------------------------------------------------
class Sammlungen07 {
   /* ---------------------------------------------------------------------
   In diesem Programm wird in eine ungeschuetzte Sammlung samsU
   vom Typ Collection<Integer> mit einer boesen Methoden ein
   Double-Objekt eingefuegt. Das gelingt ohne Ausnahme.
   Dann wird das Gleiche mit einer geschuetzten Sicht samsG auf
   samsU versucht. Dadurch wird eine Ausnahme ausgeloest.
   --------------------------------------------------------------------- */

   static public void gutAdd(Collection<Long> coLo) {
      // Zwei Objekte werden in die Sammlung coLo eingefuegt:
      coLo.add(new Long(100));
      coLo.add(new Long(200));
   } // gutAdd

   static public void erlaubtRemove(Collection<Long> coLo) {
      // Es ist erlaubt zu versuchen, aus einer Long-Sammlung ein
      // Double-Objekt zu entfernen. Normalerweise misslingt der
      // Versuch, weil die Sammlung kein solches Objekt enthaelt,
      // aber hier gelingt das Entfernen sogar:
      coLo.remove(Boese.dob);
   } // erlaubtRemove

   static public void main(String [] sonja) {

      pln("Sammlungen07: Jetzt geht es los!");
      pln("----------------------------------------");

      try {
         // Eine ungeschuetzte Sammlung samsU und eine geschuetzte Sicht
         // samsG auf diese Sammlung werden erzeugt (samsU und samsG
         // bezeichnen die selbe Sammlung, sind aber mit unterschied-
         // lichen "Sicherheitsmassnahmen" verbunden):
         Collection<Long> samsU = new ArrayList<Long>(); // Ungeschuetzt
         Collection<Long> samsG =                        //   Geschuetzt
            Collections.checkedCollection(samsU, Long.class);

         // Die ungeschuetzte Sammlung samsU wird unter anderem mit einer
         // "boesen Methode" bearbeitet. Das gelingt ohne Ausnahme:
         gutAdd(samsU);
         pln("A samsU: " + samsU);
         Boese.add(samsU);
         pln("B samsU: " + samsU);
         erlaubtRemove(samsU);
      pln("----------------------------------------");
         // Die geschuetzte Sicht samsG wird mit einer "boesen Methode"
         // bearbeitet. Dadurch wird eine Ausnahme ausgeloest:
         pln("C samsG: " + samsG);
         Boese.add(samsG);
         pln("D samsG: " + samsG);
      } catch (ClassCastException ex) {
         pln("Ausnahme: " + ex);
      }
      pln("----------------------------------------");
   } // main
   // ---------------------------------------------------------------------
   // Eine Methode mit einem kurzen Namen:
   static void pln(Object ob) {System.out.println(ob);}
   // ---------------------------------------------------------------------
} // Sammlungen07
/* ------------------------------------------------------------------------
Warnung des Compilers (betrifft nur die Klasse Boese):

D:\Java\Sammlungen07.java:35: warning: [unchecked] unchecked call to add(E)
as a member of the raw type java.util.Collection
      cr.add(dob); // Vor diesem Befehl warnt der Compiler (aber
            ^
---------------------------------------------------------------------------
Ausgabe des Programms Sammlungen07 (Zeilenumbruch "von Hand"):

Sammlungen07: Jetzt geht es los!
----------------------------------------
A samsU: [100, 200]
B samsU: [100, 200, 3.5]
----------------------------------------
C samsG: [100, 200]
Ausnahme: java.lang.ClassCastException: Attempt to insert
class java.lang.Double element into collection with
element type class java.lang.Long
----------------------------------------
------------------------------------------------------------------------ */