// Datei Sint.java
/* ------------------------------------------------------------------------
Mit Objekten der Klasse Sint ("sichere int-Werte") kann man sicher
rechnen (mit den Objektmethoden add, sub, mul, div und mod). Jedes
Sint-Objekt ist gleich dem Objekt NAN_VALUE (not a number, but a
legal Sint value) oder repraesentiert eine Ganzzahl zwischen
-Integer.MAX_VALUE und +Integer.MAX_VALUE (zum Typ Sint gehoeren
also gleich viele negative wie positive Werte und kein Sint-Wert
entspricht der negativen Zahl Integer.MIN_VALUE!).

In den folgenden Faellen ist das Ergnis einer Sint-Berechnung (mit
den Methoden add, sub etc.) gleich dem Objekt NAN_VALUE:
1. Das Ergebnis waere groesser als +Integer.MAX_VALUE oder kleiner
   als -Integer.MAX_VALUE
2. Mindestens einer der Operanden war gleich NAN_VALUE.
3. Der zweite Operand einer Division (div) oder einer Restberechnung
   (mod) ist gleich 0.

Sint-Objekte sind unveraenderbar, das Ergbnis einer Berechnung ist immer
ein neues Objekt. Ausnahme: Es gibt nur ein NAN_VALUE-Objekt und das
Ergebnis einer "falschen Berechnung" ist identisch mit diesem einen
Objekt.
------------------------------------------------------------------------ */
class Sint {
   private int wert;
   public  Sint(int wert) {this.wert = wert;}

   static final int  NAN       =  Integer.MIN_VALUE;
   static final long MIN       = -Integer.MAX_VALUE;
   static final long MAX       = +Integer.MAX_VALUE;

   static final Sint NAN_VALUE = new Sint(      NAN);
   static final Sint MIN_VALUE = new Sint((int) MIN);
   static final Sint MAX_VALUE = new Sint((int) MAX);

   // ---------------------------------------------------------------------
   public String toString() {
      if (wert==NAN) {
         return "NAN";
      } else if (wert >= 0) {
         return "+" + wert; // Ziffern mit Unterstrichen gruppieren?
      } else {
         return ""  + wert;
      }
   } // toString
   // ---------------------------------------------------------------------
   public Sint add(final Sint a) {
      if (wert==NAN || a.wert==NAN) return NAN_VALUE;
      long erg = (long) wert + (long) a.wert;
      if (erg < MIN || MAX < erg) return NAN_VALUE;
      return new Sint((int) erg);
   } // add
   // ---------------------------------------------------------------------
   public Sint sub(final Sint a) {
      if (wert==NAN || a.wert==NAN) return NAN_VALUE;
      long erg = (long) wert - (long) a.wert;
      if (erg < MIN || MAX < erg) return NAN_VALUE;
      return new Sint((int) erg);
   } // sub
   // ---------------------------------------------------------------------
   public Sint mul(final Sint a) {
      if (wert==NAN || a.wert==NAN) return NAN_VALUE;
      long erg = (long) wert * (long) a.wert;
      if (erg < MIN || MAX < erg) return NAN_VALUE;
      return new Sint((int) erg);
   } // mul
   // ---------------------------------------------------------------------
   public Sint div(final Sint a) {
      if (wert==NAN || a.wert==NAN || a.wert==0) return NAN_VALUE;
      return new Sint(wert/a.wert);
   } // div
   // ---------------------------------------------------------------------
   public Sint mod(final Sint a) {
      if (wert==NAN || a.wert==NAN || a.wert==0) return NAN_VALUE;
      return new Sint(wert%a.wert);
   } // mod
   // ---------------------------------------------------------------------
   static public void main(String[] sonja) {
      printf("Sint: Jetzt geht es los!\n");

      final int TAUSEND   = 1000;
      final int MILLION   = TAUSEND * TAUSEND;
      final int MILLIARDE = TAUSEND * MILLION;

      Sint P_0           = new Sint(0);             // "P" wie "plus"
      Sint P_1           = new Sint(1);
      Sint M_17          = new Sint(-17);           // "M" wie "minus"
      Sint M_1           = new Sint(-1);
      Sint P_1TAUSEND    = new Sint(TAUSEND);
      Sint P_1MILLION    = new Sint(MILLION);
      Sint P_1MILLIARDE  = new Sint(MILLIARDE);
      Sint P_2MILLIARDEN = new Sint(2 * MILLIARDE);

      Sint Nan01         = MAX_VALUE.add(P_1);
      Sint Nan02         = P_1.div(P_0);

      printf("---------------------------------------------\n");
      printf("P_1:                              %11s\n", P_1);
      printf("M_1:                              %11s\n", M_1);
      printf("M_17:                             %11s\n", M_17);
      printf("---------------------------------------------\n");
      printf("P_1TAUSEND:                       %11s\n", P_1TAUSEND);
      printf("P_1MILLION:                       %11s\n", P_1MILLION);
      printf("P_1MILLIARDE:                     %11s\n", P_1MILLIARDE);
      printf("P_2MILLIARDEN:                    %11s\n", P_2MILLIARDEN);
      printf("---------------------------------------------\n");
      printf("Sint.MIN_VALUE:                   %11s\n", Sint.MIN_VALUE);
      printf("Sint.MAX_VALUE:                   %11s\n", Sint.MAX_VALUE);
      printf("---------------------------------------------\n");
      printf("P_2MILLIARDEN.add(P_1TAUSEND):    %11s\n", P_2MILLIARDEN.add(P_1TAUSEND));
      printf("P_2MILLIARDEN.add(P_1MILLION):    %11s\n", P_2MILLIARDEN.add(P_1MILLION));
      printf("P_2MILLIARDEN.add(P_2MILLIARDEN): %11s\n", P_2MILLIARDEN.add(P_2MILLIARDEN));
      printf("---------------------------------------------\n");
      printf("P_1          .sub(P_2MILLIARDEN): %11s\n", P_1          .sub(P_2MILLIARDEN));
      printf("P_2MILLIARDEN.sub(P_1):           %11s\n", P_2MILLIARDEN.sub(P_1));
      printf("---------------------------------------------\n");
      printf("P_1TAUSEND   .mul(P_1MILLION):    %11s\n", P_1TAUSEND   .mul(P_1MILLION));
      printf("P_1MILLION   .mul(P_1MILLION):    %11s\n", P_1MILLION   .mul(P_1MILLION));
      printf("---------------------------------------------\n");
      printf("P_1          .div(P_0):           %11s\n", P_1          .div(P_0));
      printf("P_1          .div(P_1):           %11s\n", P_1          .div(P_1));
      printf("P_2MILLIARDEN.div(P_1MILLION):    %11s\n", P_2MILLIARDEN.div(P_1MILLION));
      printf("---------------------------------------------\n");
      printf("MAX_VALUE    .div(P_0):           %11s\n", MAX_VALUE    .div(P_0));
      printf("MAX_VALUE    .div(P_1):           %11s\n", MAX_VALUE    .div(P_1));
      printf("MAX_VALUE    .div(M_17):          %11s\n", MAX_VALUE    .div(M_17));
      printf("MAX_VALUE    .div(MIN_VALUE):     %11s\n", MAX_VALUE    .div(MIN_VALUE));
      printf("---------------------------------------------\n");
      printf("MIN_VALUE    .div(P_0):           %11s\n", MIN_VALUE    .div(P_0));
      printf("MIN_VALUE    .div(P_1):           %11s\n", MIN_VALUE    .div(P_1));
      printf("MIN_VALUE    .div(M_17):          %11s\n", MIN_VALUE    .div(M_17));
      printf("MIN_VALUE    .div(MIN_VALUE):     %11s\n", MIN_VALUE    .div(MIN_VALUE));
      printf("---------------------------------------------\n");
      printf("MAX_VALUE    .mod(P_0):           %11s\n", MAX_VALUE    .mod(P_0));
      printf("MAX_VALUE    .mod(P_1):           %11s\n", MAX_VALUE    .mod(P_1));
      printf("MAX_VALUE    .mod(M_17):          %11s\n", MAX_VALUE    .mod(M_17));
      printf("MAX_VALUE    .mod(MIN_VALUE):     %11s\n", MAX_VALUE    .mod(MIN_VALUE));
      printf("---------------------------------------------\n");
      printf("MIN_VALUE    .mod(P_0):           %11s\n", MIN_VALUE    .mod(P_0));
      printf("MIN_VALUE    .mod(P_1):           %11s\n", MIN_VALUE    .mod(P_1));
      printf("MIN_VALUE    .mod(M_17):          %11s\n", MIN_VALUE    .mod(M_17));
      printf("MIN_VALUE    .mod(MIN_VALUE):     %11s\n", MIN_VALUE    .mod(MIN_VALUE));
      printf("---------------------------------------------\n");
      printf("(Nan01 == Sint.NAN_VALUE):        %11s\n", (Nan01 == Sint.NAN_VALUE));
      printf("(Nan01 == Nan02):                 %11s\n", (Nan01 == Nan02));
      printf("---------------------------------------------\n");

      printf("Sint: Das war's erstmal!\n");
   } // main
   // ---------------------------------------------------------------------
   // Eine Methode mit einem kurzen Namen:
   static void printf(String f, Object... v) {System.out.printf(f, v);}
   // ---------------------------------------------------------------------
} // class Sint
/* ------------------------------------------------------------------------
Ausgabe des Programms Sint:

Sint: Jetzt geht es los!
---------------------------------------------
P_1:                                       +1
M_1:                                       -1
M_17:                                     -17
---------------------------------------------
P_1TAUSEND:                             +1000
P_1MILLION:                          +1000000
P_1MILLIARDE:                     +1000000000
P_2MILLIARDEN:                    +2000000000
---------------------------------------------
Sint.MIN_VALUE:                   -2147483647
Sint.MAX_VALUE:                   +2147483647
---------------------------------------------
P_2MILLIARDEN.add(P_1TAUSEND):    +2000001000
P_2MILLIARDEN.add(P_1MILLION):    +2001000000
P_2MILLIARDEN.add(P_2MILLIARDEN):         NAN
---------------------------------------------
P_1          .sub(P_2MILLIARDEN): -1999999999
P_2MILLIARDEN.sub(P_1):           +1999999999
---------------------------------------------
P_1TAUSEND   .mul(P_1MILLION):    +1000000000
P_1MILLION   .mul(P_1MILLION):            NAN
---------------------------------------------
P_1          .div(P_0):                   NAN
P_1          .div(P_1):                    +1
P_2MILLIARDEN.div(P_1MILLION):          +2000
---------------------------------------------
MAX_VALUE    .div(P_0):                   NAN
MAX_VALUE    .div(P_1):           +2147483647
MAX_VALUE    .div(M_17):           -126322567
MAX_VALUE    .div(MIN_VALUE):              -1
---------------------------------------------
MIN_VALUE    .div(P_0):                   NAN
MIN_VALUE    .div(P_1):           -2147483647
MIN_VALUE    .div(M_17):           +126322567
MIN_VALUE    .div(MIN_VALUE):              +1
---------------------------------------------
MAX_VALUE    .mod(P_0):                   NAN
MAX_VALUE    .mod(P_1):                    +0
MAX_VALUE    .mod(M_17):                   +8
MAX_VALUE    .mod(MIN_VALUE):              +0
---------------------------------------------
MIN_VALUE    .mod(P_0):                   NAN
MIN_VALUE    .mod(P_1):                    +0
MIN_VALUE    .mod(M_17):                   -8
MIN_VALUE    .mod(MIN_VALUE):              +0
---------------------------------------------
(Nan01 == Sint.NAN_VALUE):               true
(Nan01 == Nan02):                        true
---------------------------------------------
Sint: Das war's erstmal!
------------------------------------------------------------------------ */
