// Datei Vergleichen01.java
/* ------------------------------------------------------------------------
Demonstriert die Anwendungen der generischen Schnittstellen
Comparable<T> und Comparator<T>.

Wenn man die Objekte einer Klasse Mango nach verschiedenen Ordnungen
sortieren will, geht man so vor:

1. Man waehlt eine der Ordnungen als "die natuerliche Ordnung fuer
   Mango-Objekte" aus und implementiert in der Klasse Mango die
   Schnittstelle Comparable<Mango> (indem man eine der natuerlichen
   Ordnung entsprechende Methode comparTo mit einem Mango-Parameter
   programmiert).

2. Fuer jede weitere Ordnung von Mango-Objekten vereinbart man
   eine Klasse, die die Schnittstelle Comparator<Mango> implem-
   entiert, (indem man in der Klasse eine Methode compare mit
   zwei Mango-Parametern vereinbart). Von dieser Klasse laesst
   man ein (Comparator<Mango>-) Objekt erzeugen.
---------------------------------------------------------------------------
Jedes Mango-Objekt hat drei Attribute: saftmenge, guete und preis.
Mango-Objekte sollen meistens nach saftmenge, manchmal aber auch
nach guete und manchmal nach preis sortiert werden.

Deshalb implementiert die Klasse Mango die Schnittstelle
Comparable<Mango> und die comparTo-Methode definiert die
"natuerliche Ordnung" nach saft.

Ausserdem werden (hier in der Klasse Mango) zwei Comparator<Mango>-
Klassen namens NachGuete und NachPreis und zwei Comparator<Mango>-
Objekte namens nachGuete und nachPreis vereinbart, deren
compare-Methoden nach Klasse bzw. nach Preis vergleichen.

Fuegt man Mango-Objekte in eine Sammlung des Typs TreeSet<Mango>
ein, so werden sie automatisch sortiert.

Wenn man beim Erzeugen der Sammlung nichts anderes angibt, wird
sie nach der "natuerlichen Ordnung" von Mangos sortiert.
Wenn man beim Erzeugen der Sammlung das Comparator<Mango>-Objekt
nachGuete angibt, werden die Mangos darin nach ihrer Guete sortiert.
Wenn man beim Erzeugen der Sammlung das Comparator<Mango>-Objekt
nachPreis  angibt, werden die Mangos darin nach ihrem Preis sortiert.
---------------------------------------------------------------------------
Die Klassen Orange und Zitrone sind reine "Demo-Klassen" mit ziemlich
unnatuerlichen "natuerlichen Ordnungen":

Weil die Klasse Orange die Schnittstelle Comparable<Zitrone> implementiert,
kann man Orangen nur mit Zitronen (aber nicht mit Orangen) vergleichen.

Weil die Klasse Zitrone die Schnittstelle Comparable<Object> implementiert,
kann man Zitronen mit beliebigen Object-Objekten vergleichen, z.B.
mit Orangen oder Zitronen oder Mangos oder String-Objekten. Die
Vergleichsmethode compareTo ist aber nicht ganz fair, haeufig
kommt sie zu dem Ergebnis, dass "ihre" Zitrone groesser ist als
das andere Objekt.
------------------------------------------------------------------------ */
import java.util.SortedSet;  // Schnittstelle
import java.util.TreeSet;    // Klasse, implementiert SortedSet
import java.lang.Comparable; // Schnittstelle ("natuerliche Ordnung)
import java.util.Comparator; // Schnittstelle ("weitere Ordnungen)
import java.util.Arrays;     // Fuer asList
// ==============================================================
class Orange implements Comparable<Zitrone> {
   int saftmenge;

   public int compareTo(Zitrone that) {
      return this.saftmenge - that.saftmenge;
   } // compareTo

   Orange(int saftmenge) {this.saftmenge = saftmenge;}
} // class Orange
// ==============================================================
class Zitrone implements Comparable<Object>{
   int saftmenge;

   public int compareTo(Object that) {
      // Wird diese Zitrone mit einer Orange verglichen,
      // zaehlt der Saft dieser Zitrone doppelt.
      // Wird diese Zitrone mit einer Zitrone verglichen,
      // zaehlt der Saft dieser Zitrone einfach.
      // Wird diese Zitrone mit irgendeinem anderen Objekt
      // verglichen, gilt diese Zitrone als groesse (+10).
      if        (that instanceof Orange)  {
         return 2*this.saftmenge - ((Orange)  that).saftmenge;
      } else if (that instanceof Zitrone) {
         return   this.saftmenge - ((Zitrone) that).saftmenge;
      } else {
         return +10;
      }
   } // compareTo

   Zitrone(int saftmenge) {this.saftmenge = saftmenge;}
} // class Zitrone
// ==============================================================
class Mango implements Comparable<Mango> {
   int     saftmenge;
   char    guete;
   double  preis;

   public String toString() {
      return "[" + saftmenge + ", " + guete + ", " + preis + "]";
   } // toString

   public int compareTo(Mango that) {
      return this.saftmenge - that.saftmenge;
   } // compareTo

   Mango(int saftmenge, char guete, double preis) {
      this.saftmenge = saftmenge;
      this.guete     = guete;
      this.preis     = preis;
   } // Konstruktor Mango

   static class NachGuete implements Comparator<Mango> {
      public int compare(Mango m1, Mango m2) {
         return m1.guete - m2.guete;
      }
   } // class NachGuete

   static class NachPreis implements Comparator<Mango> {
      public int compare(Mango m1, Mango m2) {
         return (int) (100.0 * m1.preis - 100.0 * m2.preis);
      }
   } // class NachPreis

   static NachGuete nachGuete = new NachGuete();
   static NachPreis nachPreis = new NachPreis  ();

} // class Mango
// ==============================================================
class Vergleichen01 {
   static public void main(String[] sonja) {
      pln("Vergleichen01: Jetzt geht es los!");

      // Heringe und andere Suedfruechte:
      String  s1 = "Hering!";

      Orange  o1 = new Orange(20);
      Orange  o2 = new Orange(10);

      Zitrone z1 = new Zitrone(15);
      Zitrone z2 = new Zitrone(23);

      Mango   m1 = new Mango(10, 'B', 3.0);
      Mango   m2 = new Mango(20, 'C', 1.0);
      Mango   m3 = new Mango(30, 'A', 2.0);

      // Eine Reihung von Mangos, fuer die bequeme Initialisierung
      // der Sammlung ts01:
      Mango[] manR = {m1, m2, m3};

      // Drei TreeSet<Mango>-Sammlungen. Die erste wird natuerlich
      // sortiert und gleich mit Mangos initialisiert.
      // Fuer die anderen beiden wird erstmal nur ein
      // Comparator<Mango>-Objekt festgelegt, nachdem sie spaeter
      // sortiert werden sollen:
      SortedSet<Mango> ts01, ts02, ts03;
      ts01  = new TreeSet<Mango>(Arrays.asList(manR));
      ts02  = new TreeSet<Mango>(Mango.nachGuete);
      ts03  = new TreeSet<Mango>(Mango.nachPreis);

      // Alle Komponenten aus der Sammlung ts01 auch in die
      // Sammlungen ts02 und ts03 einfuegen (dabei werden
      // ts02 und ts03 automatisch sortiert):
      ts02.addAll(ts01);
      ts03.addAll(ts01);


      pln("------------------------------------------------");
      pln("A o1.compareTo(z1) : " + o1.compareTo(z1));
      pln("B z1.compareTo(o1) : " + z1.compareTo(o1));
      pln("C z1.compareTo(m1) : " + z1.compareTo(m1));
      pln("D z1.compareTo(s1) : " + z1.compareTo(s1));
      pln("------------------------------------------------");
//    pln("E o1.compareTo(o2) : " + o1.compareTo(o2)); // Fehler
      pln("F z1.compareTo(z2) : " + z1.compareTo(z2));
      pln("------------------------------------------------");
      pln("G m1.compareTo(m2) : " + m1.compareTo(m2));
//    pln("H m2.compareTo(o1) : " + m2.compareTo(o1)); // Fehler
      pln("---------------------------------------------------------------");
      pln("I Nach Saft:   ts01: " + ts01);
      pln("J Nach Klasse: ts02: " + ts02);
      pln("K Nach Preis:  ts03: " + ts03);
      pln("---------------------------------------------------------------");

      pln("Vergleichen01: Das war's erstmal!");
   } // main
   // ---------------------------------------------------------------------
   // Eine Methode mit einem kurzen Namen:
   static void pln(Object ob) {System.out.println(ob);}
   // ---------------------------------------------------------------------
} // class Vergleichen01
/* ------------------------------------------------------------------------
Ausgabe des Programms Vergleichen01:

Vergleichen01: Jetzt geht es los!
------------------------------------------------
A o1.compareTo(z1) : 5
B z1.compareTo(o1) : 10
C z1.compareTo(m1) : 10
D z1.compareTo(s1) : 10
------------------------------------------------
F z1.compareTo(z2) : -8
------------------------------------------------
G m1.compareTo(m2) : -10
---------------------------------------------------------------
I Nach Saft:   ts01: [[10, B, 3.0], [20, C, 1.0], [30, A, 2.0]]
J Nach Klasse: ts02: [[30, A, 2.0], [10, B, 3.0], [20, C, 1.0]]
K Nach Preis:  ts03: [[20, C, 1.0], [30, A, 2.0], [10, B, 3.0]]
---------------------------------------------------------------
Vergleichen01: Das war's erstmal!
---------------------------------------------------------------------------
Fehlermeldungen des Compilers, wenn die entsprechenden Zeilen keine
Kommentare sind:

D:\Java\Vergleichen01.java:175: compareTo(Zitrone) in Orange
cannot be applied to (Orange)
    pln("E o1.compareTo(o2): " + o1.compareTo(o2)); // Fehler
                                      ^
D:\Java\Vergleichen01.java:179: compareTo(Mango) in Mango
cannot be applied to (Orange)
    pln("H m2.compareTo(o1): " + m2.compareTo(o1)); // Fehler
                                      ^
------------------------------------------------------------------------ */
